#!/usr/bin/python
"""Simulate pytest -s -vv: discover and run some tests."""

from __future__ import print_function

import argparse
import importlib
import sys
import traceback

# Make sure we really import the correct module
# (we do not want to end up with the old "pathlib" backport to Python 2.x)
#
if sys.version_info[0] < 3:
    import pathlib2 as pathlib  # pylint: disable=import-error
else:
    import pathlib  # pylint: disable=import-error


def main():
    # type: () -> None
    """Discover tests, run them."""
    parser = argparse.ArgumentParser(
        prog="run_tests", description="Try to Python unit tests"
    )
    parser.add_argument(
        "testdir",
        nargs="?",
        type=str,
        default="unit_tests",
        help="Path to the unit tests directory",
    )
    args = parser.parse_args()

    testdir = pathlib.Path(args.testdir)
    if not testdir.parts or testdir.parts[0] in ("/", ".."):
        sys.exit("Only a subdirectory supported for the present")

    print("Looking for tests in {testdir}".format(testdir=testdir))
    collected = {}
    for fname in (path.with_suffix("") for path in testdir.rglob("test_*.py")):
        mod = importlib.import_module(".".join(fname.parts))
        for test in (name for name in dir(mod) if name.startswith("test_")):
            key = "{fname}::{test}".format(fname=fname, test=test)
            collected[key] = getattr(mod, test)

    print("Collected {count} items\n".format(count=len(collected)))
    failed = []
    for name in sorted(collected):
        print("{name}... ".format(name=name), end="")
        try:
            collected[name]()
            print("ok")
        except Exception:  # pylint: disable=broad-except
            print("FAILED")
            failed.append(sys.exc_info())

    if failed:
        print(repr(failed))
        for item in failed:
            print("\n")
            traceback.print_exception(*item)
            print("\n")
        sys.exit(1)
    print("Fine!")


if __name__ == "__main__":
    main()



# Copyright (c) 2020  Peter Pentchev <roam@ringlet.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Simulate pytest.raises() in a very naive way."""

import contextlib
import re
import sys

try:
    from types import TracebackType
    from typing import Iterator, Optional, Tuple, Type, Union  # noqa: H301

    ExceptionTuple = Union[
        Tuple[Type[BaseException], BaseException, TracebackType],
        Tuple[None, None, None],
    ]
except ImportError:
    pass


class ExceptionInfo(object):
    """Simulate the exception info returned by pytest.raises()."""

    def __init__(self, einfo=None):
        # type: (ExceptionInfo, Optional[ExceptionTuple]) -> None
        self.type = None  # type: Optional[Type[BaseException]]
        self.typename = None  # type: Optional[str]
        self.value = None  # type: Optional[BaseException]
        # pylint: disable=invalid-name
        self.tb = None  # type: Optional[TracebackType]

        if einfo is None:
            einfo = sys.exc_info()
        self.init_from(einfo)

    def init_from(self, einfo):
        # type: (ExceptionInfo, ExceptionTuple) -> None
        """Initialize the current object's members.

        This is needed because the exception is caught much later than
        the object is returned by raises().
        """
        self.type, self.value, self.tb = einfo  # pylint: disable=invalid-name
        self.typename = self.type.__name__ if self.type is not None else None
        # We do not simulate the traceback member

    def errisinstance(self, exc):
        # type: (ExceptionInfo, Type[Exception]) -> bool
        """Check whether the exception is an instance of that class."""
        return isinstance(self.value, exc)

    def match(self, regexp):
        # type: (ExceptionInfo, str) -> None
        """Check whether the exception message matches that pattern."""
        assert re.match(regexp, str(self.value))

    def __repr__(self):
        # type: (ExceptionInfo) -> str
        return "ExceptionInfo(type={etype}, value={value}, tb={tb})".format(
            etype=repr(self.type), value=repr(self.value), tb=repr(self.tb)
        )

    def __str__(self):
        # type: (ExceptionInfo) -> str
        return repr(self)


@contextlib.contextmanager
def raises(etype):
    # type: (Type[Exception]) -> Iterator[ExceptionInfo]
    """Make sure that the specified exception will be raised."""
    einfo = ExceptionInfo()
    caught = False
    try:
        yield einfo
    except etype:
        einfo.init_from(sys.exc_info())
        caught = True
    finally:
        if not caught:
            raise Exception("{etype} not thrown".format(etype=etype))



# Copyright (c) 2020  Peter Pentchev <roam@ringlet.net>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Trivial validation - when the full power of the JSON Schema is not needed.

This library provides a simplistic way to validate a dictionary against
something resembling a schema - a dictionary describing the desired data
structure by example.

The main entry point is the validate() function, but the various
validate_*() functions may be invoked directly with appropriate
arguments.

The schema used for validation is a dictionary (the top-level object must
be a dictionary). For the present, the keys may only be strings.
A special case of a dictionary with a single key "*" means any value for
a key will be accepted. Otherwise, all keys with names not starting with
a "?" character are mandatory, and any keys with names starting with
a "?" character are optional.

The dictionary values may be any of:
- a Python type signifying that the value must be an instance thereof
- a single-element list signifying that the value must be a list with
  all the elements validated by the same rules as a dictionary value
  (i.e. one of a Python type, a single-element list, a set, or
  a dictionary)
- a set signifying that the value must be exactly equal to one of
  the set elements, i.e. an enumeration of the allowed values
- a dictionary with the same semantics as described above

For example, the following schema:

    {
        "name": str,
        "id": int,
        "address": [str],
        "preferences": {
            "meal": set(("breakfast", "lunch", "brunch")),
            "colors": [{
                "name": str,
                "intensity": set(["dark", "light"])
            }]
        },
        "possessions": {
            "*": int
        }
    }

...may be used to validate the following dictionary:

    {
        "name": "A. N. Nymous",
        "id": 13,
        "address": [
            "42 Nowhere Circle",
            "Notown-at-all",
            "Unnamed territory"
        ],
        "preferences": {
            "meal": "brunch",
            "colors": [
                {"name": "blue", "intensity": "light"},
                {"name": "green", "intensity": "dark"}
            ]
        },
        "possessions": {
            "pencil": 4,
            "paper": 0
        }
    }
"""

import sys

try:
    from typing import Any, Dict, List, Tuple, Type  # noqa: H301

    SchemaType = Dict[Tuple[int, int], Dict[str, Any]]
except ImportError:
    pass


VERSION = "0.1.0"

FEATURES_STRING = "trivval=" + VERSION

FLAG_ALLOW_EXTRA = 0x0001

SCHEMA_FORMAT = {"format": {"version": {"major": int, "minor": int}}}

if sys.version_info[0] < 3:
    EQUIVALENT_TYPES = [
        (str, unicode),  # noqa: F821  # pylint: disable=undefined-variable
    ]
else:
    EQUIVALENT_TYPES = []  # type: List[Type[Any]]


class ValidationError(Exception):
    """Signal an error that occurred during the validation."""

    def __init__(self, path, err):
        # type: (ValidationError, List[str], str) -> None
        self.path = path
        self.err = err
        suffix = "" if not self.path else ": " + "/".join(path)
        super(ValidationError, self).__init__(err + suffix)


def validate_single(key, item, schema, flags):
    # type: (str, Any, Any, int) -> None
    """Validate a single dictionary value."""
    if isinstance(schema, type):
        if not isinstance(item, schema):
            for equiv in EQUIVALENT_TYPES:
                if schema in equiv and any(
                    isinstance(item, etype) for etype in equiv
                ):
                    return
            raise ValidationError(
                [key],
                "not a {t}, {vt} instead".format(
                    t=schema.__name__, vt=type(item).__name__
                ),
            )
    elif isinstance(schema, list):
        validate_list(key, item, schema, flags)
    elif isinstance(schema, set):
        if item not in schema:
            raise ValidationError([key], "not among the allowed values")
    else:
        assert isinstance(schema, dict)
        try:
            validate_dict(item, schema, flags)
        except ValidationError as err:
            raise ValidationError([key] + err.path, err.err)


def validate_list(key, value, schema, flags):
    # type: (str, List[Any], List[Any], int) -> None
    """Validate a list against a single-element schema."""
    if not isinstance(value, list):
        raise ValidationError(
            [key], "not a list, {t} instead".format(t=type(value).__name__),
        )

    assert len(schema) == 1
    for index, item in enumerate(value):
        validate_single(
            "{key}[{index}]".format(key=key, index=index),
            item,
            schema[0],
            flags,
        )


def validate_dict(value, schema, flags):
    # type: (Any, Dict[str, Any], int) -> None
    """Validate a dictionary against a schema."""
    if not isinstance(value, dict):
        raise ValidationError(
            [], "not a dictionary, {t} instead".format(t=type(value).__name__),
        )

    if len(schema.keys()) == 1 and "*" in schema:
        valtype = schema["*"]
        for key in value.keys():
            validate_single(key, value[key], valtype, flags)
        return

    extra = set(value.keys())
    for key, valtype in schema.items():
        if key.startswith("?"):
            required = False
            key = key[1:]
        else:
            required = True

        if key not in value:
            if required:
                raise ValidationError([key], "missing")
            continue
        extra.remove(key)

        validate_single(key, value[key], valtype, flags)

    if extra and not flags & FLAG_ALLOW_EXTRA:
        raise ValidationError([",".join(sorted(extra))], "extra keys")


def validate(value, schemas, flags=0):
    # type: (Dict[str, Any], SchemaType, int) -> None
    """Validate a dictionary against the appropriate schema."""
    try:
        validate_dict(value, SCHEMA_FORMAT, FLAG_ALLOW_EXTRA)
    except ValidationError:
        raise
    except Exception:
        raise ValidationError(
            ["format", "version"],
            "could not parse the version of the data format",
        )
    version = (
        value["format"]["version"]["major"],
        value["format"]["version"]["minor"],
    )
    stripped = {key: val for key, val in value.items() if key != "format"}

    same_major = sorted(
        ver
        for ver in schemas.keys()
        if ver[0] == version[0] and ver[1] <= version[1]
    )
    if not same_major:
        raise ValidationError(
            ["format", "version"], "unsupported format version"
        )

    validate_dict(stripped, schemas[same_major[-1]], flags)
